const std = @import("std");
const r11 = @import("r11.zig");

fn run_bf_tritium(code: []const u8, allocator: std.mem.Allocator) !void {
    const file = try std.fs.cwd().createFile(".code.tmp", .{});
    try file.writeAll(code);
    const proc = try std.ChildProcess.init(&.{"tritium", "-b", ".code.tmp"}, allocator);
    try proc.spawn();
    _ = proc.wait() catch {};
}

fn run_bf_fallback(code: []const u8) void {
    const stdout = std.io.getStdOut().writer();
    const stdin = std.io.getStdIn().reader();

    var tape = std.mem.zeroes([30000]u8);
    var ptr: usize = 0;
    var ip: usize = 0;

    while (ip < code.len) {
        switch (code[ip]) {
            '+' => tape[ptr] +%= 1,
            '-' => tape[ptr] -%= 1,
            '>' => ptr += 1,
            '<' => ptr -= 1,
            '.' => stdout.writeByte(tape[ptr]) catch {},
            ',' => tape[ptr] = stdin.readByte() catch tape[ptr],
            '[' => if (tape[ptr] == 0) {
                var depth: usize = 1;
                while (depth > 0) {
                    ip += 1;
                    if (code[ip] == ']') { depth -= 1; }
                    else if (code[ip] == '[') depth += 1;
                }
            },
            ']' => if (tape[ptr] != 0) {
                var depth: usize = 1;
                while (depth > 0) {
                    ip -= 1;
                    if (code[ip] == '[') { depth -= 1; }
                    else if (code[ip] == ']') depth += 1;
                }
            },
            else => {},
        }
        ip += 1;
    }
}

fn run_bf(code: []const u8, allocator: std.mem.Allocator) !void {
    const proc = try std.ChildProcess.init(&.{"tritium"}, allocator);
    proc.stdin_behavior = .Ignore;
    proc.stdout_behavior = .Ignore;
    proc.stderr_behavior = .Ignore;
    try proc.spawn();
    _ = proc.wait() catch return run_bf_fallback(code);
    return run_bf_tritium(code, allocator);
}

pub fn main() !void {
    var arena = std.heap.ArenaAllocator.init(std.heap.page_allocator);
    defer arena.deinit();
    const allocator = arena.allocator();

    var args = std.process.ArgIterator.init();
    defer args.deinit();
    _ = args.skip();

    const mode = try args.next(allocator).?;
    switch (mode[1]) {
        'c' => {
            const in = try std.fs.cwd().openFile(try args.next(allocator).?, .{});
            defer in.close();

            const contents = try in.readToEndAlloc(allocator, std.math.maxInt(u64));
            const compressed = try r11.compress(contents);

            const out = try std.fs.cwd().createFile(try args.next(allocator).?, .{});
            defer out.close();

            try out.writeAll(compressed);
        },
        'r' => {
            const file = try std.fs.cwd().openFile(try args.next(allocator).?, .{});
            defer file.close();

            const contents = try file.readToEndAlloc(allocator, std.math.maxInt(u64));
            const program = try r11.decompress(contents);

            try run_bf(program, allocator);
        },
        else => return error.UnsupportedMode,
    }
}
